@@ -63,7 +63,6 @@ namespace i2p
 		if (!port) port = SelectRandomPort ();
 		bool ipv4;  i2p::config::GetOption("ipv4", ipv4);
 		bool ipv6;  i2p::config::GetOption("ipv6", ipv6);
-		bool ssu;   i2p::config::GetOption("ssu", ssu);
 		bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
 		bool ssu2;  i2p::config::GetOption("ssu2.enabled", ssu2);
 		bool ygg;   i2p::config::GetOption("meshnets.yggdrasil", ygg);
@@ -109,19 +108,14 @@ namespace i2p
 					routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv);
 				}
 			}
-			if (ssu)
-			{
-				routerInfo.AddSSUAddress (host.c_str(), port, nullptr);
-				caps |= i2p::data::RouterInfo::eReachable; // R
-			}
 			if (ssu2)
 			{
 				if (ssu2Published)
 				{
 					uint16_t ssu2Port; i2p::config::GetOption ("ssu2.port", ssu2Port);
-					if (!ssu2Port) ssu2Port = ssu ? (port + 1) : port;
+					if (!ssu2Port) ssu2Port = port;
 					routerInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, boost::asio::ip::address_v4::from_string (host), ssu2Port);
-				}	
+				}
 				else
 				{
 					addressCaps |= i2p::data::RouterInfo::AddressCaps::eV4;
@@ -131,7 +125,7 @@ namespace i2p
 		}
 		if (ipv6)
 		{
-			std::string host = "::1";
+			std::string host;
 			if (!i2p::config::IsDefault("host") && !ipv4) // override if v6 only
 				i2p::config::GetOption("host", host);
 			else
@@ -142,6 +136,7 @@ namespace i2p
 
 			if (ntcp2)
 			{
+				bool added = false;
 				if (ntcp2Published)
 				{
 					std::string ntcp2Host;
@@ -149,29 +144,33 @@ namespace i2p
 						i2p::config::GetOption ("ntcp2.addressv6", ntcp2Host);
 					else
 						ntcp2Host = host;
-					routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address_v6::from_string (ntcp2Host), port);
+					if (!ntcp2Host.empty () && port)
+					{
+						routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, boost::asio::ip::address_v6::from_string (ntcp2Host), port);
+						added = true;
+					}
 				}
-				else
+				if (!added)
 				{
 					if (!ipv4) // no other ntcp2 addresses yet
 						routerInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv);
 					addressCaps |= i2p::data::RouterInfo::AddressCaps::eV6;
 				}
 			}
-			if (ssu)
-			{
-				routerInfo.AddSSUAddress (host.c_str(), port, nullptr);
-				caps |= i2p::data::RouterInfo::eReachable; // R
-			}
 			if (ssu2)
 			{
+				bool added = false;
 				if (ssu2Published)
 				{
 					uint16_t ssu2Port; i2p::config::GetOption ("ssu2.port", ssu2Port);
-					if (!ssu2Port) ssu2Port = ssu ? (port + 1) : port;
-					routerInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, boost::asio::ip::address_v6::from_string (host), ssu2Port);
-				}	
-				else
+					if (!ssu2Port) ssu2Port = port;
+					if (!host.empty () && ssu2Port)
+					{
+						routerInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, boost::asio::ip::address_v6::from_string (host), ssu2Port);
+						added = true;
+					}
+				}
+				if (!added)
 				{
 					if (!ipv4) // no other ssu2 addresses yet
 						routerInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro);
@@ -201,8 +200,8 @@ namespace i2p
 		uint16_t port = rand () % (30777 - 9111) + 9111; // I2P network ports range
 		if (port == 9150) port = 9151; // Tor browser
 		return port;
-	}	
-		
+	}
+
 	void RouterContext::UpdateRouterInfo ()
 	{
 		m_RouterInfo.CreateBuffer (m_Keys);
@@ -236,13 +235,6 @@ namespace i2p
 		fk.write ((char *)m_SSU2Keys.get (), sizeof (SSU2PrivateKeys));
 	}
 
-	bool RouterContext::IsSSU2Only () const
-	{
-		auto transports = m_RouterInfo.GetCompatibleTransports (false);
-		return (transports & (i2p::data::RouterInfo::eSSU2V4 | i2p::data::RouterInfo::eSSU2V6)) &&
-			!(transports & (i2p::data::RouterInfo::eSSUV4 | i2p::data::RouterInfo::eSSUV6));
-	}
-
 	void RouterContext::SetStatus (RouterStatus status)
 	{
 		if (status != m_Status)
@@ -263,12 +255,6 @@ namespace i2p
 		}
 	}
 
-	void RouterContext::SetStatusSSU2 (RouterStatus status)
-	{
-		if (IsSSU2Only ())
-			SetStatus (status);
-	}
-
 	void RouterContext::SetStatusV6 (RouterStatus status)
 	{
 		if (status != m_StatusV6)
@@ -289,18 +275,14 @@ namespace i2p
 		}
 	}
 
-	void RouterContext::SetStatusV6SSU2 (RouterStatus status)
-	{
-		if (IsSSU2Only ())
-			SetStatusV6 (status);
-	}
-
 	void RouterContext::UpdatePort (int port)
 	{
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
 		bool updated = false;
-		for (auto& address : m_RouterInfo.GetAddresses ())
+		for (auto& address : *addresses)
 		{
-			if (address->port != port && (address->transportStyle == i2p::data::RouterInfo::eTransportSSU || IsSSU2Only ()))
+			if (address && address->port != port && address->transportStyle == i2p::data::RouterInfo::eTransportSSU2)
 			{
 				address->port = port;
 				updated = true;
@@ -313,10 +295,12 @@ namespace i2p
 	void RouterContext::PublishNTCP2Address (int port, bool publish, bool v4, bool v6, bool ygg)
 	{
 		if (!m_NTCP2Keys) return;
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
 		bool updated = false;
-		for (auto& address : m_RouterInfo.GetAddresses ())
+		for (auto& address : *addresses)
 		{
-			if (address->IsNTCP2 () && (address->port != port || address->published != publish))
+			if (address && address->IsNTCP2 () && (address->port != port || address->published != publish))
 			{
 				bool isAddr = v4 && address->IsV4 ();
 				if (!isAddr && (v6 || ygg))
@@ -342,25 +326,23 @@ namespace i2p
 
 	void RouterContext::UpdateNTCP2Address (bool enable)
 	{
-		auto& addresses = m_RouterInfo.GetAddresses ();
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
 		bool found = false, updated = false;
-		for (auto it = addresses.begin (); it != addresses.end ();)
+		for (auto& it: *addresses)
 		{
-			if ((*it)->IsNTCP2 ())
+			if (it && it->IsNTCP2 ())
 			{
 				found = true;
 				if (enable)
 				{
-					(*it)->s = m_NTCP2Keys->staticPublicKey;
-					memcpy ((*it)->i, m_NTCP2Keys->iv, 16);
-					it++;
-				}	
+					it->s = m_NTCP2Keys->staticPublicKey;
+					memcpy (it->i, m_NTCP2Keys->iv, 16);
+				}
 				else
-					it = addresses.erase (it);
+					it.reset ();
 				updated = true;
 			}
-			else
-				it++;
 		}
 		if (enable && !found)
 		{
@@ -374,21 +356,23 @@ namespace i2p
 	void RouterContext::PublishSSU2Address (int port, bool publish, bool v4, bool v6)
 	{
 		if (!m_SSU2Keys) return;
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
 		int newPort = 0;
 		if (!port)
 		{
-			for (const auto& address : m_RouterInfo.GetAddresses ())
-				if (address->port)
+			for (const auto& address : *addresses)
+				if (address && address->port)
 				{
 					newPort = address->port;
 					break;
-				}	
+				}
 			if (!newPort) newPort = SelectRandomPort ();
-		}	
+		}
 		bool updated = false;
-		for (auto& address : m_RouterInfo.GetAddresses ())
+		for (auto& address : *addresses)
 		{
-			if (address->IsSSU2 () && (!address->port || address->port != port || address->published != publish) &&
+			if (address && address->IsSSU2 () && (!address->port || address->port != port || address->published != publish) &&
 				((v4 && address->IsV4 ()) || (v6 && address->IsV6 ())))
 			{
 				if (port) address->port = port;
@@ -407,25 +391,23 @@ namespace i2p
 
 	void RouterContext::UpdateSSU2Address (bool enable)
 	{
-		auto& addresses = m_RouterInfo.GetAddresses ();
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
 		bool found = false, updated = false;
-		for (auto it = addresses.begin (); it != addresses.end ();)
+		for (auto& it : *addresses)
 		{
-			if ((*it)->IsSSU2 ())
+			if (it && it->IsSSU2 ())
 			{
 				found = true;
 				if (enable)
 				{
-					(*it)->s = m_SSU2Keys->staticPublicKey;
-					(*it)->i = m_SSU2Keys->intro;
-					it++;
-				}	
-				else	
-					it = addresses.erase (it);
+					it->s = m_SSU2Keys->staticPublicKey;
+					it->i = m_SSU2Keys->intro;
+				}
+				else
+					it.reset ();
 				updated = true;
 			}
-			else
-				it++;
 		}
 		if (enable && !found)
 		{
@@ -438,12 +420,12 @@ namespace i2p
 				if (ipv6) m_RouterInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, i2p::data::RouterInfo::AddressCaps::eV6);
 			}
 			else
-			{	
+			{
 				uint8_t addressCaps = 0;
 				if (ipv4) addressCaps |= i2p::data::RouterInfo::AddressCaps::eV4;
 				if (ipv6) addressCaps |= i2p::data::RouterInfo::AddressCaps::eV6;
 				m_RouterInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, addressCaps);
-			}	
+			}
 			updated = true;
 		}
 		if (updated)
@@ -452,59 +434,63 @@ namespace i2p
 
 	void RouterContext::UpdateAddress (const boost::asio::ip::address& host)
 	{
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
 		bool updated = false;
-		for (auto& address : m_RouterInfo.GetAddresses ())
+		if (host.is_v4 ())
 		{
-			if (address->host != host && address->IsCompatible (host) &&
-				!i2p::util::net::IsYggdrasilAddress (address->host))
+			auto addr = (*addresses)[i2p::data::RouterInfo::eNTCP2V4Idx];
+			if (addr && addr->host != host)
 			{
-				// update host
-				address->host = host;
+				addr->host = host;
 				updated = true;
 			}
-			if (host.is_v6 () && address->IsV6 () && address->ssu && 
-			    (!address->ssu->mtu || updated))
+			addr = (*addresses)[i2p::data::RouterInfo::eSSU2V4Idx];
+			if (addr && addr->host != host)
 			{
-				// update MTU
-				auto mtu = i2p::util::net::GetMTU (host);
-				if (mtu)
-				{
-					LogPrint (eLogDebug, "Router: Our v6 MTU=", mtu);
-					int maxMTU = i2p::util::net::GetMaxMTU (host.to_v6 ());
-					if (mtu > maxMTU) 
-					{ 
-						mtu = maxMTU;
-						LogPrint(eLogWarning, "Router: MTU dropped to upper limit of ", maxMTU, " bytes");
-					}
-					if (mtu && !address->IsSSU2 ()) // SSU1
-						mtu = (mtu >> 4) << 4; // round to multiple of 16
-					address->ssu->mtu = mtu;
-					updated = true;
-				}
+				addr->host = host;
+				updated = true;
 			}
 		}
+		else if (host.is_v6 ())
+		{
+			auto addr = (*addresses)[i2p::data::RouterInfo::eNTCP2V6Idx];
+			if (addr && addr->host != host)
+			{
+				addr->host = host;
+				updated = true;
+			}
+			addr = (*addresses)[i2p::data::RouterInfo::eSSU2V6Idx];
+			if (addr && (addr->host != host || !addr->ssu->mtu))
+			{
+				addr->host = host;
+				if (m_StatusV6 != eRouterStatusProxy)
+				{
+					// update MTU
+					auto mtu = i2p::util::net::GetMTU (host);
+					if (mtu)
+					{
+						LogPrint (eLogDebug, "Router: Our v6 MTU=", mtu);
+						int maxMTU = i2p::util::net::GetMaxMTU (host.to_v6 ());
+						if (mtu > maxMTU)
+						{
+							mtu = maxMTU;
+							LogPrint(eLogWarning, "Router: MTU dropped to upper limit of ", maxMTU, " bytes");
+						}
+						addr->ssu->mtu = mtu;
+					}
+				}
+				updated = true;
+			}
+		}
+
 		auto ts = i2p::util::GetSecondsSinceEpoch ();
 		if (updated || ts > m_LastUpdateTime + ROUTER_INFO_UPDATE_INTERVAL)
 			UpdateRouterInfo ();
 	}
 
-	bool RouterContext::AddIntroducer (const i2p::data::RouterInfo::Introducer& introducer)
-	{
-		bool ret = m_RouterInfo.AddIntroducer (introducer);
-		if (ret)
-			UpdateRouterInfo ();
-		return ret;
-	}
-
-	void RouterContext::RemoveIntroducer (const boost::asio::ip::udp::endpoint& e)
-	{
-		if (m_RouterInfo.RemoveIntroducer (e))
-			UpdateRouterInfo ();
-	}
-
 	bool RouterContext::AddSSU2Introducer (const i2p::data::RouterInfo::Introducer& introducer, bool v4)
 	{
-		if (!IsSSU2Only ()) return false;
 		bool ret = m_RouterInfo.AddSSU2Introducer (introducer, v4);
 		if (ret)
 			UpdateRouterInfo ();
@@ -513,26 +499,20 @@ namespace i2p
 
 	void RouterContext::RemoveSSU2Introducer (const i2p::data::IdentHash& h, bool v4)
 	{
-		if (!IsSSU2Only ()) return;
 		if (m_RouterInfo.RemoveSSU2Introducer (h, v4))
 			UpdateRouterInfo ();
 	}
 
 	void RouterContext::ClearSSU2Introducers (bool v4)
 	{
-		bool updated = false;
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto& addr : addresses)
-			if (addr->IsSSU2 () && ((v4 && addr->IsV4 ()) || (!v4 && addr->IsV6 ())) &&
-			    addr->ssu && !addr->ssu->introducers.empty ())
-			{
-				addr->ssu->introducers.clear ();
-				updated = true;
-			}
-		if (updated)
+		auto addr = m_RouterInfo.GetSSU2Address (v4);
+		if (addr && !addr->ssu->introducers.empty ())
+		{
+			addr->ssu->introducers.clear ();
 			UpdateRouterInfo ();
-	}	
-		
+		}
+	}
+
 	void RouterContext::SetFloodfill (bool floodfill)
 	{
 		m_IsFloodfill = floodfill;
@@ -631,50 +611,6 @@ namespace i2p
 		return m_RouterInfo.GetCaps () & i2p::data::RouterInfo::eUnreachable;
 	}
 
-	void RouterContext::RemoveNTCPAddress (bool v4only)
-	{
-		bool updated = false;
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto it = addresses.begin (); it != addresses.end ();)
-		{
-			if ((*it)->transportStyle == i2p::data::RouterInfo::eTransportNTCP && !(*it)->IsNTCP2 () &&
-				(!v4only || (*it)->host.is_v4 ()))
-			{
-				it = addresses.erase (it);
-				updated = true;
-				if (v4only) break; // otherwise might be more than one address
-			}
-			else
-				++it;
-		}
-		if (updated)
-			m_RouterInfo.UpdateSupportedTransports ();
-	}
-
-	void RouterContext::RemoveSSUAddress ()
-	{
-		bool updated = false;
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto it = addresses.begin (); it != addresses.end ();)
-		{
-			if ((*it)->transportStyle == i2p::data::RouterInfo::eTransportSSU)
-			{	
-				it = addresses.erase (it);
-				updated = true;
-			}	
-			else
-				++it;
-		}
-		if (updated)
-			m_RouterInfo.UpdateSupportedTransports ();
-	}
-		
-	void RouterContext::SetUnreachableSSU2 (bool v4, bool v6)
-	{
-		if (IsSSU2Only ())
-			SetUnreachable (v4, v6);
-	}
-
 	void RouterContext::SetUnreachable (bool v4, bool v6)
 	{
 		if (v4 || (v6 && !SupportsV4 ()))
@@ -689,16 +625,18 @@ namespace i2p
 		}
 		uint16_t port = 0;
 		// delete previous introducers
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto& addr : addresses)
-			if (addr->ssu && (!addr->IsSSU2 () || IsSSU2Only ()) && 
-			    ((v4 && addr->IsV4 ()) || (v6 && addr->IsV6 ())))
-			{
-				addr->published = false;
-				addr->caps &= ~i2p::data::RouterInfo::eSSUIntroducer; // can't be introducer
-				addr->ssu->introducers.clear ();
-				port = addr->port;
-			}
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (addresses)
+		{
+			for (auto& addr : *addresses)
+				if (addr && addr->ssu && ((v4 && addr->IsV4 ()) || (v6 && addr->IsV6 ())))
+				{
+					addr->published = false;
+					addr->caps &= ~i2p::data::RouterInfo::eSSUIntroducer; // can't be introducer
+					addr->ssu->introducers.clear ();
+					port = addr->port;
+				}
+		}
 		// unpublish NTCP2 addreeses
 		bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
 		if (ntcp2)
@@ -722,20 +660,19 @@ namespace i2p
 		}
 		uint16_t port = 0;
 		// delete previous introducers
-		bool isSSU2Published = IsSSU2Only (); // TODO
-		if (isSSU2Published)
-			i2p::config::GetOption ("ssu2.published", isSSU2Published);
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto& addr : addresses)
-			if (addr->ssu && (!addr->IsSSU2 () || isSSU2Published) && 
-			    ((v4 && addr->IsV4 ()) || (v6 && addr->IsV6 ())))
-			{
-				addr->published = true;
-				addr->caps |= i2p::data::RouterInfo::eSSUIntroducer;
-				addr->ssu->introducers.clear ();
-				if (addr->port && (!addr->IsSSU2 () || IsSSU2Only ()))
-					port = addr->port;
-			}
+		bool isSSU2Published; i2p::config::GetOption ("ssu2.published", isSSU2Published);
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (addresses)
+		{
+			for (auto& addr : *addresses)
+				if (addr && addr->ssu && isSSU2Published && ((v4 && addr->IsV4 ()) || (v6 && addr->IsV6 ())))
+				{
+					addr->published = true;
+					addr->caps |= i2p::data::RouterInfo::eSSUIntroducer;
+					addr->ssu->introducers.clear ();
+					if (addr->port) port = addr->port;
+				}
+		}
 		// publish NTCP2
 		bool ntcp2; i2p::config::GetOption("ntcp2.enabled", ntcp2);
 		if (ntcp2)
@@ -758,40 +695,33 @@ namespace i2p
 		if (supportsV6)
 		{
 			// insert v6 addresses if necessary
-			bool foundSSU = false, foundNTCP2 = false, foundSSU2 = false;
+			bool foundNTCP2 = false, foundSSU2 = false;
 			uint16_t port = 0;
-			auto& addresses = m_RouterInfo.GetAddresses ();
-			for (auto& addr: addresses)
+			auto addresses = m_RouterInfo.GetAddresses ();
+			if (addresses)
 			{
-				if (addr->IsV6 () && !i2p::util::net::IsYggdrasilAddress (addr->host))
+				for (auto& addr: *addresses)
 				{
-					switch (addr->transportStyle)
+					if (addr && addr->IsV6 () && !i2p::util::net::IsYggdrasilAddress (addr->host))
 					{
-						case i2p::data::RouterInfo::eTransportSSU:
-							foundSSU = true;
-						break;
-						case i2p::data::RouterInfo::eTransportNTCP:
-							foundNTCP2 = true;
-						break;
-						case i2p::data::RouterInfo::eTransportSSU2:
-							foundSSU2 = true;
-						break;
-						default: ;
+						switch (addr->transportStyle)
+						{
+							case i2p::data::RouterInfo::eTransportNTCP2:
+								foundNTCP2 = true;
+							break;
+							case i2p::data::RouterInfo::eTransportSSU2:
+								foundSSU2 = true;
+							break;
+							default: ;
+						}
 					}
+					if (addr) port = addr->port;
 				}
-				port = addr->port;
 			}
-			if (!port) 
-			{	
+			if (!port)
+			{
 				i2p::config::GetOption("port", port);
 				if (!port) port = SelectRandomPort ();
-			}	
-			// SSU
-			bool ssu; i2p::config::GetOption("ssu", ssu);
-			if (!foundSSU && ssu)
-			{
-				std::string host = "::1"; // TODO: read host
-				m_RouterInfo.AddSSUAddress (host.c_str (), port, nullptr);
 			}
 			// NTCP2
 			if (!foundNTCP2)
@@ -825,7 +755,7 @@ namespace i2p
 					if (ssu2Published)
 					{
 						uint16_t ssu2Port; i2p::config::GetOption ("ssu2.port", ssu2Port);
-						if (!ssu2Port) ssu2Port = ssu ? (port + 1) : port;
+						if (!ssu2Port) ssu2Port = port;
 						m_RouterInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, boost::asio::ip::address::from_string ("::1"), ssu2Port);
 					}
 					else
@@ -847,40 +777,35 @@ namespace i2p
 		// update
 		if (supportsV4)
 		{
-			bool foundSSU = false, foundNTCP2 = false, foundSSU2 = false;
+			bool foundNTCP2 = false, foundSSU2 = false;
 			std::string host = "127.0.0.1";
 			uint16_t port = 0;
-			auto& addresses = m_RouterInfo.GetAddresses ();
-			for (auto& addr: addresses)
+			auto addresses = m_RouterInfo.GetAddresses ();
+			if (addresses)
 			{
-				if (addr->IsV4 ())
+				for (auto& addr: *addresses)
 				{
-					switch (addr->transportStyle)
+					if (addr && addr->IsV4 ())
 					{
-						case i2p::data::RouterInfo::eTransportSSU:
-							foundSSU = true;
-						break;
-						case i2p::data::RouterInfo::eTransportNTCP:
-							foundNTCP2 = true;
-						break;
-						case i2p::data::RouterInfo::eTransportSSU2:
-							foundSSU2 = true;
-						break;
-						default: ;
+						switch (addr->transportStyle)
+						{
+							case i2p::data::RouterInfo::eTransportNTCP2:
+								foundNTCP2 = true;
+							break;
+							case i2p::data::RouterInfo::eTransportSSU2:
+								foundSSU2 = true;
+							break;
+							default: ;
+						}
 					}
+					if (addr && addr->port) port = addr->port;
 				}
-				if (addr->port) port = addr->port;
 			}
-			if (!port) 
-			{	
+			if (!port)
+			{
 				i2p::config::GetOption("port", port);
 				if (!port) port = SelectRandomPort ();
-			}	
-			// SSU
-			bool ssu; i2p::config::GetOption("ssu", ssu);
-			if (!foundSSU && ssu)
-				m_RouterInfo.AddSSUAddress (host.c_str (), port, nullptr);
-
+			}
 			// NTCP2
 			if (!foundNTCP2)
 			{
@@ -908,7 +833,7 @@ namespace i2p
 					if (ssu2Published)
 					{
 						uint16_t ssu2Port; i2p::config::GetOption ("ssu2.port", ssu2Port);
-						if (!ssu2Port) ssu2Port = ssu ? (port + 1) : port;
+						if (!ssu2Port) ssu2Port = port;
 						m_RouterInfo.AddSSU2Address (m_SSU2Keys->staticPublicKey, m_SSU2Keys->intro, boost::asio::ip::address::from_string ("127.0.0.1"), ssu2Port);
 					}
 					else
@@ -926,23 +851,26 @@ namespace i2p
 	{
 		if (supportsmesh)
 		{
+			auto addresses = m_RouterInfo.GetAddresses ();
+			if (!addresses) return;
 			m_RouterInfo.EnableMesh ();
+			if ((*addresses)[i2p::data::RouterInfo::eNTCP2V6MeshIdx]) return; // we have mesh address already
 			uint16_t port = 0;
 			i2p::config::GetOption ("ntcp2.port", port);
 			if (!port) i2p::config::GetOption("port", port);
-			bool foundMesh = false;
-			auto& addresses = m_RouterInfo.GetAddresses ();
-			for (auto& addr: addresses)
+			if (!port)
 			{
-				if (!port) port = addr->port;
-				if (i2p::util::net::IsYggdrasilAddress (addr->host))
+				for (auto& addr: *addresses)
 				{
-					foundMesh = true;
-					break;
+					if (addr && addr->port)
+					{
+						port = addr->port;
+						break;
+					}
 				}
 			}
-			if (!foundMesh)
-				m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, host, port);
+			if (!port) port = SelectRandomPort ();
+			m_RouterInfo.AddNTCP2Address (m_NTCP2Keys->staticPublicKey, m_NTCP2Keys->iv, host, port);
 		}
 		else
 			m_RouterInfo.DisableMesh ();
@@ -952,36 +880,14 @@ namespace i2p
 	void RouterContext::SetMTU (int mtu, bool v4)
 	{
 		if (mtu < 1280 || mtu > 1500) return;
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto& addr: addresses)
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
+		for (auto& addr: *addresses)
 		{
-			if (addr->ssu && ((v4 && addr->IsV4 ()) || (!v4 && addr->IsV6 ())))
+			if (addr && addr->ssu && ((v4 && addr->IsV4 ()) || (!v4 && addr->IsV6 ())))
 			{
-				if (!addr->IsSSU2 ()) // SSU1
-				{
-					// round to multiple of 16
-					if (v4)
-					{
-						if (mtu > 1484) mtu = 1484;
-						else
-						{
-							mtu -= 12;
-							mtu = (mtu >> 4) << 4;
-							mtu += 12;
-						}
-					}
-					else
-					{
-						if (mtu > 1488) mtu = 1488;
-						else
-							mtu = (mtu >> 4) << 4;
-					}
-				}
-				if (mtu)
-				{
-					addr->ssu->mtu = mtu;
-					LogPrint (eLogDebug, "Router: MTU for ", v4 ? "ipv4" : "ipv6", " address ", addr->host.to_string(), " is set to ", mtu);
-				}
+				addr->ssu->mtu = mtu;
+				LogPrint (eLogDebug, "Router: MTU for ", v4 ? "ipv4" : "ipv6", " address ", addr->host.to_string(), " is set to ", mtu);
 			}
 		}
 	}
@@ -990,10 +896,11 @@ namespace i2p
 	{
 		bool isYgg = i2p::util::net::IsYggdrasilAddress (host);
 		bool updated = false;
-		auto& addresses = m_RouterInfo.GetAddresses ();
-		for (auto& addr: addresses)
+		auto addresses = m_RouterInfo.GetAddresses ();
+		if (!addresses) return;
+		for (auto& addr: *addresses)
 		{
-			if (addr->IsPublishedNTCP2 ())
+			if (addr && addr->IsPublishedNTCP2 ())
 			{
 				bool isYgg1 = i2p::util::net::IsYggdrasilAddress (addr->host);
 				if (addr->IsV6 () && ((isYgg && isYgg1) || (!isYgg && !isYgg1)))
