@@ -206,26 +206,24 @@ namespace data
 		s.read ((char *)&m_Timestamp, sizeof (m_Timestamp));
 		m_Timestamp = be64toh (m_Timestamp);
 		// read addresses
-		auto addresses = boost::make_shared<Addresses>();
+		auto addresses = netdb.NewRouterInfoAddresses ();
 		uint8_t numAddresses;
 		s.read ((char *)&numAddresses, sizeof (numAddresses));
-		addresses->reserve (numAddresses);
 		for (int i = 0; i < numAddresses; i++)
 		{
 			uint8_t supportedTransports = 0;
-			auto address = std::make_shared<Address> ();
+			auto address = netdb.NewRouterInfoAddress ();
 			uint8_t cost; // ignore
 			s.read ((char *)&cost, sizeof (cost));
 			s.read ((char *)&address->date, sizeof (address->date));
-			bool isHost = false, isIntroKey = false, isStaticKey = false, isV2 = false;
-			Tag<32> iV2; // for 'i' field in SSU, TODO: remove later
+			bool isHost = false, isStaticKey = false, isV2 = false;
 			char transportStyle[6];
 			ReadString (transportStyle, 6, s);
 			if (!strncmp (transportStyle, "NTCP", 4)) // NTCP or NTCP2
-				address->transportStyle = eTransportNTCP;
+				address->transportStyle = eTransportNTCP2;
 			else if (!strncmp (transportStyle, "SSU", 3)) // SSU or SSU2
 			{
-				address->transportStyle = (transportStyle[3] == '2') ? eTransportSSU2 : eTransportSSU;
+				address->transportStyle = eTransportSSU2;
 				address->ssu.reset (new SSUExt ());
 				address->ssu->mtu = 0;
 			}
@@ -257,20 +255,31 @@ namespace data
 					if (!ecode && !address->host.is_unspecified ()) isHost = true;
 				}
 				else if (!strcmp (key, "port"))
-					address->port = boost::lexical_cast<int>(value);
+				{
+					try
+					{
+						address->port = boost::lexical_cast<int>(value);
+					}
+					catch (std::exception& ex)
+					{
+						LogPrint (eLogWarning, "RouterInfo: 'port' exception ", ex.what ());
+					}
+				}
 				else if (!strcmp (key, "mtu"))
 				{
 					if (address->ssu)
-						address->ssu->mtu = boost::lexical_cast<int>(value);
+					{
+						try
+						{
+							address->ssu->mtu = boost::lexical_cast<int>(value);
+						}
+						catch (std::exception& ex)
+						{
+							LogPrint (eLogWarning, "RouterInfo: 'mtu' exception ", ex.what ());
+						}
+					}
 					else
-						LogPrint (eLogWarning, "RouterInfo: Unexpected field 'mtu' for NTCP");
-				}
-				else if (!strcmp (key, "key"))
-				{
-					if (address->ssu)
-						isIntroKey = (Base64ToByteStream (value, strlen (value), address->i, 32) == 32);
-					else
-						LogPrint (eLogWarning, "RouterInfo: Unexpected field 'key' for NTCP");
+						LogPrint (eLogWarning, "RouterInfo: Unexpected field 'mtu' for NTCP2");
 				}
 				else if (!strcmp (key, "caps"))
 					address->caps = ExtractAddressCaps (value);
@@ -288,8 +297,6 @@ namespace data
 					}
 					else if (address->IsSSU2 ())
 						Base64ToByteStream (value, strlen (value), address->i, 32);
-					else
-						Base64ToByteStream (value, strlen (value), iV2, 32);
 				}
 				else if (!strcmp (key, "v"))
 				{
@@ -322,22 +329,40 @@ namespace data
 					}
 					Introducer& introducer = address->ssu->introducers.at (index);
 					if (!strcmp (key, "ihost"))
-					{
-						boost::system::error_code ecode;
-						introducer.iHost = boost::asio::ip::address::from_string (value, ecode);
-					}
+						introducer.isH = false; // SSU1
 					else if (!strcmp (key, "iport"))
-						introducer.iPort = boost::lexical_cast<int>(value);
+						introducer.isH = false; // SSU1
 					else if (!strcmp (key, "itag"))
-						introducer.iTag = boost::lexical_cast<uint32_t>(value);
-					else if (!strcmp (key, "ikey") || !strcmp (key, "ih"))
-						Base64ToByteStream (value, strlen (value), introducer.iKey, 32);
+					{
+						try
+						{
+							introducer.iTag = boost::lexical_cast<uint32_t>(value);
+						}
+						catch (std::exception& ex)
+						{
+							LogPrint (eLogWarning, "RouterInfo: 'itag' exception ", ex.what ());
+						}
+					}
+					else if (!strcmp (key, "ih"))
+					{
+						Base64ToByteStream (value, strlen (value), introducer.iH, 32);
+						introducer.isH = true;
+					}
 					else if (!strcmp (key, "iexp"))
-						introducer.iExp = boost::lexical_cast<uint32_t>(value);
+					{
+						try
+						{
+							introducer.iExp = boost::lexical_cast<uint32_t>(value);
+						}
+						catch (std::exception& ex)
+						{
+							LogPrint (eLogWarning, "RouterInfo: 'iexp' exception ", ex.what ());
+						}
+					}
 				}
 				if (!s) return;
 			}
-			if (address->transportStyle == eTransportNTCP)
+			if (address->transportStyle == eTransportNTCP2)
 			{
 				if (isStaticKey)
 				{
@@ -361,49 +386,7 @@ namespace data
 					}
 				}
 			}
-			else if (address->transportStyle == eTransportSSU)
-			{
-				if (isIntroKey)
-				{
-					if (isHost)
-						supportedTransports |= address->host.is_v4 () ? eSSUV4 : eSSUV6;
-					else if (address->caps & AddressCaps::eV6)
-					{
-						supportedTransports |= eSSUV6;
-						if (address->caps & AddressCaps::eV4) supportedTransports |= eSSUV4; // in additional to v6
-					}
-					else
-						supportedTransports |= eSSUV4; // in case if host or 6 caps is not preasented, we assume 4
-					if (address->ssu && !address->ssu->introducers.empty ())
-					{
-						// exclude invalid introducers
-						uint32_t ts = i2p::util::GetSecondsSinceEpoch ();
-						int numValid = 0;
-						for (auto& it: address->ssu->introducers)
-						{
-							if (!it.iExp) it.iExp = m_Timestamp/1000 + NETDB_INTRODUCEE_EXPIRATION_TIMEOUT;
-							if (ts <= it.iExp && it.iPort > 0 &&
-								((it.iHost.is_v4 () && address->IsV4 ()) || (it.iHost.is_v6 () && address->IsV6 ())))
-								numValid++;
-							else
-							{	
-								it.iPort = 0;
-								if (isV2) numValid++;
-							}	
-						}
-						if (numValid)
-							m_ReachableTransports |= supportedTransports;
-						else
-							address->ssu->introducers.resize (0);
-					}
-					else if (isHost && address->port)
-					{
-						address->published = true;
-						m_ReachableTransports |= supportedTransports;
-					}
-				}
-			}
-			if (address->transportStyle == eTransportSSU2 || (isV2 && address->transportStyle == eTransportSSU))
+			else if (address->transportStyle == eTransportSSU2 && isV2)
 			{
 				if (address->IsV4 ()) supportedTransports |= eSSU2V4;
 				if (address->IsV6 ()) supportedTransports |= eSSU2V6;
@@ -411,57 +394,38 @@ namespace data
 				{
 					if (address->host.is_v4 ()) m_ReachableTransports |= eSSU2V4;
 					if (address->host.is_v6 ()) m_ReachableTransports |= eSSU2V6;
+					address->published = true;
 				}
-				if (address->transportStyle == eTransportSSU2)
+				if (address->ssu && !address->ssu->introducers.empty ())
 				{
-					if (address->port) address->published = true;
-					if (address->ssu && !address->ssu->introducers.empty ())
-					{
-						// exclude invalid introducers
-						uint32_t ts = i2p::util::GetSecondsSinceEpoch ();
-						int numValid = 0;
-						for (auto& it: address->ssu->introducers)
-						{
-							if (it.iTag && ts <= it.iExp)						
-								numValid++;
-							else	
-								it.iTag = 0;
-						}
-						if (numValid)
-							m_ReachableTransports |= supportedTransports;
-						else
-							address->ssu->introducers.resize (0);
-					}	
-				}
-				else
-				{
-					// create additional SSU2 address. TODO: remove later
-					auto ssu2addr = std::make_shared<Address> ();
-					ssu2addr->transportStyle = eTransportSSU2;
-					ssu2addr->host = address->host; ssu2addr->port = address->port;
-					ssu2addr->s = address->s; ssu2addr->i = iV2;
-					ssu2addr->date = address->date; ssu2addr->caps = address->caps;
-					ssu2addr->published = address->published;
-					ssu2addr->ssu.reset (new SSUExt ()); ssu2addr->ssu->mtu = address->ssu->mtu;
+					// exclude invalid introducers
 					uint32_t ts = i2p::util::GetSecondsSinceEpoch ();
-					if (!address->ssu->introducers.empty ())
-					{	
-						for (const auto& introducer: address->ssu->introducers)
-							if (!introducer.iPort && introducer.iHost.is_unspecified () && ts < introducer.iExp) // SSU2
-								ssu2addr->ssu->introducers.push_back (introducer);
-						if (!ssu2addr->ssu->introducers.empty ())
-							m_ReachableTransports |= supportedTransports;
-					}	
-					addresses->push_back(ssu2addr);
-				}	
+					int numValid = 0;
+					for (auto& it: address->ssu->introducers)
+					{
+						if (it.iTag && ts < it.iExp && it.isH)
+							numValid++;
+						else
+							it.iTag = 0;
+					}
+					if (numValid)
+						m_ReachableTransports |= supportedTransports;
+					else
+						address->ssu->introducers.resize (0);
+				}
 			}
 			if (supportedTransports)
 			{
 				if (!(m_SupportedTransports & supportedTransports)) // avoid duplicates
-					addresses->push_back(address);
+				{
+					for (uint8_t i = 0; i < eNumTransports; i++)
+						if ((1 << i) & supportedTransports)
+							(*addresses)[i] = address;
+				}
 				m_SupportedTransports |= supportedTransports;
 			}
 		}
+		// update addresses
 #if (BOOST_VERSION >= 105300)
 		boost::atomic_store (&m_Addresses, addresses);
 #else
@@ -590,10 +554,10 @@ namespace data
 				case CAPS_FLAG_V6:
 					caps |= AddressCaps::eV6;
 				break;
-				case CAPS_FLAG_SSU_TESTING:
+				case CAPS_FLAG_SSU2_TESTING:
 					caps |= AddressCaps::eSSUTesting;
 				break;
-				case CAPS_FLAG_SSU_INTRODUCER:
+				case CAPS_FLAG_SSU2_INTRODUCER:
 					caps |= AddressCaps::eSSUIntroducer;
 				break;
 				default: ;
@@ -656,36 +620,13 @@ namespace data
 		return l+1;
 	}
 
-
-	void RouterInfo::AddSSUAddress (const char * host, int port, const uint8_t * key, int mtu)
-	{
-		auto addr = std::make_shared<Address>();
-		addr->host = boost::asio::ip::address::from_string (host);
-		addr->port = port;
-		addr->transportStyle = eTransportSSU;
-		addr->published = true;
-		addr->caps = i2p::data::RouterInfo::eSSUTesting | i2p::data::RouterInfo::eSSUIntroducer; // BC;
-		addr->date = 0;
-		addr->ssu.reset (new SSUExt ());
-		addr->ssu->mtu = mtu;
-		if (key)
-			memcpy (addr->i, key, 32);
-		else
-			RAND_bytes (addr->i, 32);
-		for (const auto& it: *m_Addresses) // don't insert same address twice
-			if (*it == *addr) return;
-		m_SupportedTransports |= addr->host.is_v6 () ? eSSUV6 : eSSUV4;
-		m_ReachableTransports |= addr->host.is_v6 () ? eSSUV6 : eSSUV4;
-		m_Addresses->push_back(std::move(addr));
-	}
-
 	void RouterInfo::AddNTCP2Address (const uint8_t * staticKey, const uint8_t * iv,
 		const boost::asio::ip::address& host, int port, uint8_t caps)
 	{
 		auto addr = std::make_shared<Address>();
 		addr->host = host;
 		addr->port = port;
-		addr->transportStyle = eTransportNTCP;
+		addr->transportStyle = eTransportNTCP2;
 		addr->caps = caps;
 		addr->date = 0;
 		if (port) addr->published = true;
@@ -695,13 +636,23 @@ namespace data
 		{
 			m_SupportedTransports |= eNTCP2V4;
 			if (addr->published) m_ReachableTransports |= eNTCP2V4;
+			(*m_Addresses)[eNTCP2V4Idx] = addr;
 		}
 		if (addr->IsV6 ())
 		{
-			m_SupportedTransports |= eNTCP2V6;
-			if (addr->published) m_ReachableTransports |= eNTCP2V6;
+			if (i2p::util::net::IsYggdrasilAddress (addr->host))
+			{
+				m_SupportedTransports |= eNTCP2V6Mesh;
+				m_ReachableTransports |= eNTCP2V6Mesh;
+				(*m_Addresses)[eNTCP2V6MeshIdx] = addr;
+			}
+			else
+			{
+				m_SupportedTransports |= eNTCP2V6;
+				if (addr->published) m_ReachableTransports |= eNTCP2V6;
+				(*m_Addresses)[eNTCP2V6Idx] = addr;
+			}
 		}
-		m_Addresses->push_back(std::move(addr));
 	}
 
 	void RouterInfo::AddSSU2Address (const uint8_t * staticKey, const uint8_t * introKey, uint8_t caps)
@@ -715,9 +666,16 @@ namespace data
 		addr->ssu->mtu = 0;
 		memcpy (addr->s, staticKey, 32);
 		memcpy (addr->i, introKey, 32);
-		if (addr->IsV4 ()) m_SupportedTransports |= eSSU2V4;
-		if (addr->IsV6 ()) m_SupportedTransports |= eSSU2V6;
-		m_Addresses->push_back(std::move(addr));
+		if (addr->IsV4 ())
+		{
+			m_SupportedTransports |= eSSU2V4;
+			(*m_Addresses)[eSSU2V4Idx] = addr;
+		}
+		if (addr->IsV6 ())
+		{
+			m_SupportedTransports |= eSSU2V6;
+			(*m_Addresses)[eSSU2V6Idx] = addr;
+		}
 	}
 
 	void RouterInfo::AddSSU2Address (const uint8_t * staticKey, const uint8_t * introKey,
@@ -738,58 +696,14 @@ namespace data
 		{
 			m_SupportedTransports |= eSSU2V4;
 			m_ReachableTransports |= eSSU2V4;
+			(*m_Addresses)[eSSU2V4Idx] = addr;
 		}
 		if (addr->IsV6 ())
 		{
 			m_SupportedTransports |= eSSU2V6;
 			m_ReachableTransports |= eSSU2V6;
+			(*m_Addresses)[eSSU2V6Idx] = addr;
 		}
-		m_Addresses->push_back(std::move(addr));
-	}
-
-	bool RouterInfo::AddIntroducer (const Introducer& introducer)
-	{
-		for (auto& addr : *m_Addresses)
-		{
-			if (addr->transportStyle == eTransportSSU &&
-				((addr->IsV4 () && introducer.iHost.is_v4 ()) || (addr->IsV6 () && introducer.iHost.is_v6 ())))
-			{
-				for (auto& intro: addr->ssu->introducers)
-					if (intro.iTag == introducer.iTag) return false; // already presented
-				addr->ssu->introducers.push_back (introducer);
-				m_ReachableTransports |= (addr->IsV4 () ? eSSUV4 : eSSUV6);
-				return true;
-			}
-		}
-		return false;
-	}
-
-	bool RouterInfo::RemoveIntroducer (const boost::asio::ip::udp::endpoint& e)
-	{
-		for (auto& addr: *m_Addresses)
-		{
-			if (addr->transportStyle == eTransportSSU &&
-				((addr->IsV4 () && e.address ().is_v4 ()) || (addr->IsV6 () && e.address ().is_v6 ())))
-			{
-				for (auto it = addr->ssu->introducers.begin (); it != addr->ssu->introducers.end (); ++it)
-					if (boost::asio::ip::udp::endpoint (it->iHost, it->iPort) == e)
-					{
-						addr->ssu->introducers.erase (it);
-						if (addr->ssu->introducers.empty ())
-							m_ReachableTransports &= ~(addr->IsV4 () ? eSSUV4 : eSSUV6);
-						return true;
-					}
-			}
-		}
-		return false;
-	}
-
-	bool RouterInfo::IsSSU (bool v4only) const
-	{
-		if (v4only)
-			return m_SupportedTransports & eSSUV4;
-		else
-			return m_SupportedTransports & (eSSUV4 | eSSUV6);
 	}
 
 	bool RouterInfo::IsNTCP2 (bool v4only) const
@@ -828,21 +742,17 @@ namespace data
 	{
 		if (IsV6 ())
 		{
-			for (auto it = m_Addresses->begin (); it != m_Addresses->end ();)
+			if ((*m_Addresses)[eNTCP2V6Idx])
 			{
-				auto addr = *it;
-				if (addr->IsV6 ())
-				{
-					if (addr->IsV4 ())
-					{
-						addr->caps &= ~AddressCaps::eV6;
-						++it;
-					}
-					else
-						it = m_Addresses->erase (it);
-				}
-				else
-					++it;
+				if ((*m_Addresses)[eNTCP2V6Idx]->IsV4 ())
+					(*m_Addresses)[eNTCP2V6Idx]->caps &= ~AddressCaps::eV6;
+				(*m_Addresses)[eNTCP2V6Idx].reset ();
+			}
+			if ((*m_Addresses)[eSSU2V6Idx])
+			{
+				if ((*m_Addresses)[eSSU2V6Idx]->IsV4 ())
+					(*m_Addresses)[eSSU2V6Idx]->caps &= ~AddressCaps::eV6;
+				(*m_Addresses)[eSSU2V6Idx].reset ();
 			}
 			UpdateSupportedTransports ();
 		}
@@ -852,21 +762,17 @@ namespace data
 	{
 		if (IsV4 ())
 		{
-			for (auto it = m_Addresses->begin (); it != m_Addresses->end ();)
+			if ((*m_Addresses)[eNTCP2V4Idx])
 			{
-				auto addr = *it;
-				if (addr->IsV4 ())
-				{
-					if (addr->IsV6 ())
-					{
-						addr->caps &= ~AddressCaps::eV4;
-						++it;
-					}
-					else
-						it = m_Addresses->erase (it);
-				}
-				else
-					++it;
+				if ((*m_Addresses)[eNTCP2V4Idx]->IsV6 ())
+					(*m_Addresses)[eNTCP2V4Idx]->caps &= ~AddressCaps::eV4;
+				(*m_Addresses)[eNTCP2V4Idx].reset ();
+			}
+			if ((*m_Addresses)[eSSU2V4Idx])
+			{
+				if ((*m_Addresses)[eSSU2V4Idx]->IsV6 ())
+					(*m_Addresses)[eSSU2V4Idx]->caps &= ~AddressCaps::eV4;
+				(*m_Addresses)[eSSU2V4Idx].reset ();
 			}
 			UpdateSupportedTransports ();
 		}
@@ -887,57 +793,24 @@ namespace data
 		{
 			m_SupportedTransports &= ~eNTCP2V6Mesh;
 			m_ReachableTransports &= ~eNTCP2V6Mesh;
-			for (auto it = m_Addresses->begin (); it != m_Addresses->end ();)
-			{
-				auto addr = *it;
-				if (i2p::util::net::IsYggdrasilAddress (addr->host))
-					it = m_Addresses->erase (it);
-				else
-					++it;
-			}
+			(*m_Addresses)[eNTCP2V6MeshIdx].reset ();
 		}
 	}
 
-	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSUAddress (bool v4only) const
-	{
-		return GetAddress (
-			[v4only](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportSSU) && (!v4only || address->IsV4 ());
-			});
-	}
-
-	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSUV6Address () const
-	{
-		return GetAddress (
-			[](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportSSU) && address->IsV6();
-			});
-	}
-
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSU2V4Address () const
 	{
-		return GetAddress (
-			[](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportSSU2) && address->IsV4();
-			});
+		return (*GetAddresses ())[eSSU2V4Idx];
 	}
 
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSU2V6Address () const
 	{
-		return GetAddress (
-			[](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportSSU2) && address->IsV6();
-			});
+		return (*GetAddresses ())[eSSU2V6Idx];
 	}
 
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSU2Address (bool v4) const
 	{
 		if (v4)
-		{	
+		{
 			if (m_SupportedTransports & eSSU2V4)
 				return GetSSU2V4Address ();
 		}
@@ -945,10 +818,19 @@ namespace data
 		{
 			if (m_SupportedTransports & eSSU2V6)
 				return GetSSU2V6Address ();
-		}	
+		}
 		return nullptr;
-	}	
-		
+	}
+
+	boost::shared_ptr<RouterInfo::Addresses> RouterInfo::GetAddresses () const
+	{
+#if (BOOST_VERSION >= 105300)
+		return boost::atomic_load (&m_Addresses);
+#else
+		return m_Addresses;
+#endif
+	}
+
 	template<typename Filter>
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetAddress (Filter filter) const
 	{
@@ -959,7 +841,7 @@ namespace data
 		auto addresses = m_Addresses;
 #endif
 		for (const auto& address : *addresses)
-			if (filter (address)) return address;
+			if (address && filter (address)) return address;
 
 		return nullptr;
 	}
@@ -977,40 +859,29 @@ namespace data
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetSSU2AddressWithStaticKey (const uint8_t * key, bool isV6) const
 	{
 		if (!key) return nullptr;
-		return GetAddress (
-			[key, isV6](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return address->IsSSU2 () && !memcmp (address->s, key, 32) && 
-					((isV6 && address->IsV6 ()) || (!isV6 && address->IsV4 ()));
-			});
+		auto addr = (*GetAddresses ())[isV6 ? eSSU2V6Idx : eSSU2V4Idx];
+		if (addr && !memcmp (addr->s, key, 32))
+		    return addr;
+		return nullptr;
 	}
 
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetPublishedNTCP2V4Address () const
 	{
-		return GetAddress (
-			[](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return address->IsPublishedNTCP2 () && address->host.is_v4 ();
-			});
+		auto addr = (*GetAddresses ())[eNTCP2V4Idx];
+		if (addr && addr->IsPublishedNTCP2 ()) return addr;
+		return nullptr;
 	}
 
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetPublishedNTCP2V6Address () const
 	{
-		return GetAddress (
-			[](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return address->IsPublishedNTCP2 () && address->host.is_v6 () &&
-					!i2p::util::net::IsYggdrasilAddress (address->host);
-			});
+		auto addr = (*GetAddresses ())[eNTCP2V6Idx];
+		if (addr && addr->IsPublishedNTCP2 ()) return addr;
+		return nullptr;
 	}
 
 	std::shared_ptr<const RouterInfo::Address> RouterInfo::GetYggdrasilAddress () const
 	{
-		return GetAddress (
-			[](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return address->IsPublishedNTCP2 () && i2p::util::net::IsYggdrasilAddress (address->host);
-			});
+		return (*GetAddresses ())[eNTCP2V6MeshIdx];
 	}
 
 	std::shared_ptr<RouterProfile> RouterInfo::GetProfile () const
@@ -1030,60 +901,29 @@ namespace data
 	bool RouterInfo::IsEligibleFloodfill () const
 	{
 		// floodfill must be reachable by ipv4, >= 0.9.38 and not DSA
-		return IsReachableBy (eNTCP2V4 | eSSUV4) && m_Version >= NETDB_MIN_FLOODFILL_VERSION &&
+		return IsReachableBy (eNTCP2V4 | eSSU2V4) && m_Version >= NETDB_MIN_FLOODFILL_VERSION &&
 			GetIdentity ()->GetSigningKeyType () != SIGNING_KEY_TYPE_DSA_SHA1;
 	}
 
-	bool RouterInfo::IsPeerTesting (bool v4) const
-	{
-		if (!(m_SupportedTransports & (v4 ? eSSUV4 : eSSUV6))) return false;
-		return (bool)GetAddress (
-			[v4](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportSSU) && address->IsPeerTesting () &&
-					((v4 && address->IsV4 ()) || (!v4 && address->IsV6 ())) && address->IsReachableSSU ();
-			});
-	}
-
 	bool RouterInfo::IsSSU2PeerTesting (bool v4) const
 	{
 		if (!(m_SupportedTransports & (v4 ? eSSU2V4 : eSSU2V6))) return false;
-		return (bool)GetAddress (
-			[v4](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->IsSSU2 ()) && address->IsPeerTesting () &&
-					((v4 && address->IsV4 ()) || (!v4 && address->IsV6 ())) && address->IsReachableSSU ();
-			});
-	}	
-		
-	bool RouterInfo::IsIntroducer (bool v4) const
-	{
-		if (!(m_SupportedTransports & (v4 ? eSSUV4 : eSSUV6))) return false;
-		return (bool)GetAddress (
-			[v4](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->transportStyle == eTransportSSU) && address->IsIntroducer () &&
-					((v4 && address->IsV4 ()) || (!v4 && address->IsV6 ())) && !address->host.is_unspecified ();
-			});
+		auto addr = (*GetAddresses ())[v4 ? eSSU2V4Idx : eSSU2V6Idx];
+		return addr && addr->IsPeerTesting () && addr->IsReachableSSU ();
 	}
 
 	bool RouterInfo::IsSSU2Introducer (bool v4) const
 	{
 		if (!(m_SupportedTransports & (v4 ? eSSU2V4 : eSSU2V6))) return false;
-		return (bool)GetAddress (
-			[v4](std::shared_ptr<const RouterInfo::Address> address)->bool
-			{
-				return (address->IsSSU2 ()) && address->IsIntroducer () &&
-					((v4 && address->IsV4 ()) || (!v4 && address->IsV6 ())) && !address->host.is_unspecified ();
-			});
+		auto addr = (*GetAddresses ())[v4 ? eSSU2V4Idx : eSSU2V6Idx];
+		return addr && addr->IsIntroducer () && !addr->host.is_unspecified () && addr->port;
 	}
-		
+
 	void RouterInfo::SetUnreachableAddressesTransportCaps (uint8_t transports)
 	{
 		for (auto& addr: *m_Addresses)
 		{
-			// TODO: implement SSU
-			if (!addr->published && (addr->transportStyle == eTransportNTCP || addr->transportStyle == eTransportSSU2))
+			if (addr && !addr->published)
 			{
 				addr->caps &= ~(eV4 | eV6);
 				addr->caps |= transports;
@@ -1097,30 +937,25 @@ namespace data
 		m_ReachableTransports = 0;
 		for (const auto& addr: *m_Addresses)
 		{
+			if (!addr) continue;
 			uint8_t transports = 0;
 			switch (addr->transportStyle)
-			{        
-				case eTransportNTCP:
+			{
+				case eTransportNTCP2:
 					if (addr->IsV4 ()) transports |= eNTCP2V4;
 					if (addr->IsV6 ())
 						transports |= (i2p::util::net::IsYggdrasilAddress (addr->host) ? eNTCP2V6Mesh : eNTCP2V6);
 					if (addr->IsPublishedNTCP2 ())
 						m_ReachableTransports |= transports;
 				break;
-				case eTransportSSU:
-					if (addr->IsV4 ()) transports |= eSSUV4;
-					if (addr->IsV6 ()) transports |= eSSUV6;
-					if (addr->IsReachableSSU ())
-						m_ReachableTransports |= transports;
-				break;
 				case eTransportSSU2:
 					if (addr->IsV4 ()) transports |= eSSU2V4;
 					if (addr->IsV6 ()) transports |= eSSU2V6;
 					if (addr->IsReachableSSU ())
 						m_ReachableTransports |= transports;
-				break;	
-				default: ;	
-			}	
+				break;
+				default: ;
+			}
 			m_SupportedTransports |= transports;
 		}
 	}
@@ -1200,29 +1035,42 @@ namespace data
 
 	void LocalRouterInfo::WriteToStream (std::ostream& s) const
 	{
+		auto addresses = GetAddresses ();
+		if (!addresses) return;
+
 		uint64_t ts = htobe64 (GetTimestamp ());
 		s.write ((const char *)&ts, sizeof (ts));
-
 		// addresses
-		const Addresses& addresses = GetAddresses ();
-		uint8_t numAddresses = addresses.size ();
-		s.write ((char *)&numAddresses, sizeof (numAddresses));
-		for (const auto& addr_ptr : addresses)
+		uint8_t numAddresses = 0;
+		for (size_t idx = 0; idx < addresses->size(); idx++)
 		{
+			auto addr_ptr = (*addresses)[idx];
+			if (!addr_ptr) continue;
+			if (idx == eNTCP2V6Idx && addr_ptr == (*addresses)[eNTCP2V4Idx]) continue;
+			if (idx == eSSU2V6Idx && addr_ptr == (*addresses)[eSSU2V4Idx]) continue;
+			numAddresses++;
+		}
+		s.write ((char *)&numAddresses, sizeof (numAddresses));
+		for (size_t idx = 0; idx < addresses->size(); idx++)
+		{
+			auto addr_ptr = (*addresses)[idx];
+			if (!addr_ptr) continue;
+			if (idx == eNTCP2V6Idx && addr_ptr == (*addresses)[eNTCP2V4Idx]) continue;
+			if (idx == eSSU2V6Idx && addr_ptr == (*addresses)[eSSU2V4Idx]) continue;
 			const Address& address = *addr_ptr;
 			// calculate cost
 			uint8_t cost = 0x7f;
-			if (address.transportStyle == eTransportNTCP)
+			if (address.transportStyle == eTransportNTCP2)
 				cost = address.published ? COST_NTCP2_PUBLISHED : COST_NTCP2_NON_PUBLISHED;
-			else if (address.transportStyle == eTransportSSU)
-				cost = address.published ? COST_SSU_DIRECT : COST_SSU_THROUGH_INTRODUCERS;
 			else if (address.transportStyle == eTransportSSU2)
 				cost = address.published ? COST_SSU2_DIRECT : COST_SSU2_NON_PUBLISHED;
+			else
+				continue; // skip unknown address
 			s.write ((const char *)&cost, sizeof (cost));
 			s.write ((const char *)&address.date, sizeof (address.date));
 			std::stringstream properties;
 			bool isPublished = false;
-			if (address.transportStyle == eTransportNTCP)
+			if (address.transportStyle == eTransportNTCP2)
 			{
 				if (address.IsNTCP2 ())
 				{
@@ -1244,43 +1092,6 @@ namespace data
 				else
 					continue; // don't write NTCP address
 			}
-			else if (address.transportStyle == eTransportSSU)
-			{
-				WriteString ("SSU", s);
-				// caps
-				WriteString ("caps", properties);
-				properties << '=';
-				std::string caps;
-				if (address.IsPeerTesting ()) caps += CAPS_FLAG_SSU_TESTING;
-				if (address.host.is_v4 ())
-				{
-					if (address.published)
-					{
-						isPublished = true;
-						if (address.IsIntroducer ()) caps += CAPS_FLAG_SSU_INTRODUCER;
-					}
-					else
-						caps += CAPS_FLAG_V4;
-				}
-				else if (address.host.is_v6 ())
-				{
-					if (address.published)
-					{
-						isPublished = true;
-						if (address.IsIntroducer ()) caps += CAPS_FLAG_SSU_INTRODUCER;
-					}
-					else
-						caps += CAPS_FLAG_V6;
-				}
-				else
-				{
-					if (address.IsV4 ()) caps += CAPS_FLAG_V4;
-					if (address.IsV6 ()) caps += CAPS_FLAG_V6;
-					if (caps.empty ()) caps += CAPS_FLAG_V4;
-				}
-				WriteString (caps, properties);
-				properties << ';';
-			}
 			else if (address.transportStyle == eTransportSSU2)
 			{
 				WriteString ("SSU2", s);
@@ -1289,8 +1100,8 @@ namespace data
 				if (address.published)
 				{
 					isPublished = true;
-					if (address.IsPeerTesting ()) caps += CAPS_FLAG_SSU_TESTING;
-					if (address.IsIntroducer ()) caps += CAPS_FLAG_SSU_INTRODUCER;
+					if (address.IsPeerTesting ()) caps += CAPS_FLAG_SSU2_TESTING;
+					if (address.IsIntroducer ()) caps += CAPS_FLAG_SSU2_INTRODUCER;
 				}
 				else
 				{
@@ -1323,7 +1134,7 @@ namespace data
 				size_t len = address.IsSSU2 () ? 32 : 16;
 				WriteString (address.i.ToBase64 (len), properties); properties << ';';
 			}
-			if (address.transportStyle == eTransportSSU || address.IsSSU2 ())
+			if (address.transportStyle == eTransportSSU2)
 			{
 				// write introducers if any
 				if (address.ssu && !address.ssu->introducers.empty())
@@ -1340,45 +1151,18 @@ namespace data
 						}
 						i++;
 					}
-					if (address.transportStyle == eTransportSSU)
-					{
-						i = 0;
-						for (const auto& introducer: address.ssu->introducers)
-						{
-							WriteString ("ihost" + boost::lexical_cast<std::string>(i), properties);
-							properties << '=';
-							WriteString (introducer.iHost.to_string (), properties);
-							properties << ';';
-							i++;
-						}
-					}
 					i = 0;
 					for (const auto& introducer: address.ssu->introducers)
 					{
-						if (address.IsSSU2 ())
-							WriteString ("ih" + boost::lexical_cast<std::string>(i), properties);
-						else
-							WriteString ("ikey" + boost::lexical_cast<std::string>(i), properties);
+						WriteString ("ih" + boost::lexical_cast<std::string>(i), properties);
 						properties << '=';
 						char value[64];
-						size_t l = ByteStreamToBase64 (introducer.iKey, 32, value, 64);
+						size_t l = ByteStreamToBase64 (introducer.iH, 32, value, 64);
 						value[l] = 0;
 						WriteString (value, properties);
 						properties << ';';
 						i++;
 					}
-					if (address.transportStyle == eTransportSSU)
-					{
-						i = 0;
-						for (const auto& introducer: address.ssu->introducers)
-						{
-							WriteString ("iport" + boost::lexical_cast<std::string>(i), properties);
-							properties << '=';
-							WriteString (boost::lexical_cast<std::string>(introducer.iPort), properties);
-							properties << ';';
-							i++;
-						}
-					}
 					i = 0;
 					for (const auto& introducer: address.ssu->introducers)
 					{
@@ -1390,18 +1174,8 @@ namespace data
 					}
 				}
 			}
-			if (address.transportStyle == eTransportSSU)
-			{
-				// write intro key
-				WriteString ("key", properties);
-				properties << '=';
-				char value[64];
-				size_t l = ByteStreamToBase64 (address.i, 32, value, 64);
-				value[l] = 0;
-				WriteString (value, properties);
-				properties << ';';
-			}
-			if (address.transportStyle == eTransportSSU || address.IsSSU2 ())
+
+			if (address.transportStyle == eTransportSSU2)
 			{
 				// write mtu
 				if (address.ssu && address.ssu->mtu)
@@ -1412,7 +1186,7 @@ namespace data
 					properties << ';';
 				}
 			}
-			if ((isPublished || (address.ssu && !address.IsSSU2 ())) && address.port)
+			if (isPublished && address.port)
 			{
 				WriteString ("port", properties);
 				properties << '=';
@@ -1483,37 +1257,37 @@ namespace data
 
 	bool LocalRouterInfo::AddSSU2Introducer (const Introducer& introducer, bool v4)
 	{
-		for (auto& addr : GetAddresses ())
+		auto addresses = GetAddresses ();
+		if (!addresses) return false;
+		auto addr = (*addresses)[v4 ? eSSU2V4Idx : eSSU2V6Idx];
+		if (addr)
 		{
-			if (addr->IsSSU2 () && ((v4 && addr->IsV4 ()) || (!v4 && addr->IsV6 ())))
-			{
-				for (auto& intro: addr->ssu->introducers)
-					if (intro.iTag == introducer.iTag) return false; // already presented
-				addr->ssu->introducers.push_back (introducer);
-				SetReachableTransports (GetReachableTransports () | ((addr->IsV4 () ? eSSU2V4 : eSSU2V6)));
-				return true;
-			}
+			for (auto& intro: addr->ssu->introducers)
+				if (intro.iTag == introducer.iTag) return false; // already presented
+			addr->ssu->introducers.push_back (introducer);
+			SetReachableTransports (GetReachableTransports () | ((addr->IsV4 () ? eSSU2V4 : eSSU2V6)));
+			return true;
 		}
 		return false;
-	}	
+	}
 
 	bool LocalRouterInfo::RemoveSSU2Introducer (const IdentHash& h, bool v4)
 	{
-		for (auto& addr: GetAddresses ())
+		auto addresses = GetAddresses ();
+		if (!addresses) return false;
+		auto addr = (*addresses)[v4 ? eSSU2V4Idx : eSSU2V6Idx];
+		if (addr)
 		{
-			if (addr->IsSSU2 () && ((v4 && addr->IsV4 ()) || (!v4 && addr->IsV6 ())))
-			{
-				for (auto it = addr->ssu->introducers.begin (); it != addr->ssu->introducers.end (); ++it)
-					if (h == it->iKey)
-					{
-						addr->ssu->introducers.erase (it);
-						if (addr->ssu->introducers.empty ())
-							SetReachableTransports (GetReachableTransports () & ~(addr->IsV4 () ? eSSU2V4 : eSSU2V6));
-						return true;
-					}
-			}
+			for (auto it = addr->ssu->introducers.begin (); it != addr->ssu->introducers.end (); ++it)
+				if (h == it->iH)
+				{
+					addr->ssu->introducers.erase (it);
+					if (addr->ssu->introducers.empty ())
+						SetReachableTransports (GetReachableTransports () & ~(addr->IsV4 () ? eSSU2V4 : eSSU2V6));
+					return true;
+				}
 		}
 		return false;
-	}	
+	}
 }
 }
