@@ -435,7 +435,7 @@ namespace stream
 					LogPrint(eLogError, "Streaming: Packet ", seqn, "sent from the future, sendTime=", sentPacket->sendTime);
 					rtt = 1;
 				}
-				m_RTT = (m_RTT*seqn + rtt)/(seqn + 1);
+				m_RTT = std::round ((m_RTT*seqn + rtt)/(seqn + 1.0));
 				m_RTO = m_RTT*1.5; // TODO: implement it better
 				LogPrint (eLogDebug, "Streaming: Packet ", seqn, " acknowledged rtt=", rtt, " sentTime=", sentPacket->sendTime);
 				m_SentPackets.erase (it++);
@@ -474,6 +474,48 @@ namespace stream
 			Close (); // check is all outgoing messages have been sent and we can send close
 	}
 
+	size_t Stream::Receive (uint8_t * buf, size_t len, int timeout)
+	{
+		if (!len) return 0;
+		size_t ret = 0;
+		volatile bool done = false;
+		std::condition_variable newDataReceived;
+		std::mutex newDataReceivedMutex;
+		AsyncReceive (boost::asio::buffer (buf, len),
+			[&ret, &done, &newDataReceived, &newDataReceivedMutex](const boost::system::error_code& ecode, std::size_t bytes_transferred)
+			{
+				if (ecode == boost::asio::error::timed_out)
+					ret = 0;
+				else
+					ret = bytes_transferred;
+				std::unique_lock<std::mutex> l(newDataReceivedMutex);
+				newDataReceived.notify_all ();
+				done = true;
+			},
+			timeout);
+		if (!done)
+		{	std::unique_lock<std::mutex> l(newDataReceivedMutex);
+			if (!done && newDataReceived.wait_for (l, std::chrono::seconds (timeout)) == std::cv_status::timeout)
+				ret = 0;
+		}
+		if (!done)
+		{
+			// make sure that AsycReceive complete
+			auto s = shared_from_this();
+			m_Service.post ([s]()
+		    {
+				s->m_ReceiveTimer.cancel ();
+			});
+			int i = 0;
+			while (!done && i < 100) // 1 sec
+			{
+				std::this_thread::sleep_for (std::chrono::milliseconds(10));
+				i++;
+			}
+		}
+		return ret;
+	}
+
 	size_t Stream::Send (const uint8_t * buf, size_t len)
 	{
 		AsyncSend (buf, len, nullptr);
@@ -729,7 +771,7 @@ namespace stream
 				Terminate ();
 			break;
 			default:
-				LogPrint (eLogWarning, "Streaming: Unexpected stream status ", (int)m_Status, "sSID=", m_SendStreamID);
+				LogPrint (eLogWarning, "Streaming: Unexpected stream status=", (int)m_Status, " for sSID=", m_SendStreamID);
 		};
 	}
 
@@ -855,7 +897,7 @@ namespace stream
 			for (const auto& it: packets)
 			{
 				auto msg = m_RoutingSession->WrapSingleMessage (m_LocalDestination.CreateDataMessage (
-					it->GetBuffer (), it->GetLength (), m_Port, !m_RoutingSession->IsRatchets ()));
+					it->GetBuffer (), it->GetLength (), m_Port, !m_RoutingSession->IsRatchets (), it->IsSYN ()));
 				msgs.push_back (i2p::tunnel::TunnelMessageBlock
 					{
 						i2p::tunnel::eDeliveryTypeTunnel,
@@ -1085,8 +1127,6 @@ namespace stream
 		m_Owner (owner), m_LocalPort (localPort), m_Gzip (gzip),
 		m_PendingIncomingTimer (m_Owner->GetService ())
 	{
-		if (m_Gzip)
-			m_Deflator.reset (new i2p::data::GzipDeflator);
 	}
 
 	StreamingDestination::~StreamingDestination ()
@@ -1341,6 +1381,26 @@ namespace stream
 		acceptor (stream);
 	}
 
+	std::shared_ptr<Stream> StreamingDestination::AcceptStream (int timeout)
+	{
+		std::shared_ptr<i2p::stream::Stream> stream;
+		std::condition_variable streamAccept;
+		std::mutex streamAcceptMutex;
+		std::unique_lock<std::mutex> l(streamAcceptMutex);
+		AcceptOnce (
+			[&streamAccept, &streamAcceptMutex, &stream](std::shared_ptr<i2p::stream::Stream> s)
+		    {
+				stream = s;
+				std::unique_lock<std::mutex> l(streamAcceptMutex);
+				streamAccept.notify_all ();
+			});
+		if (timeout)
+			streamAccept.wait_for (l, std::chrono::seconds (timeout));
+		else
+			streamAccept.wait (l);
+		return stream;
+	}
+
 	void StreamingDestination::HandlePendingIncomingTimer (const boost::system::error_code& ecode)
 	{
 		if (ecode != boost::asio::error::operation_aborted)
@@ -1365,7 +1425,7 @@ namespace stream
 	}
 
 	std::shared_ptr<I2NPMessage> StreamingDestination::CreateDataMessage (
-		const uint8_t * payload, size_t len, uint16_t toPort, bool checksum)
+		const uint8_t * payload, size_t len, uint16_t toPort, bool checksum, bool gzip)
 	{
 		size_t size;
 		auto msg = m_I2NPMsgsPool.AcquireShared ();
@@ -1373,8 +1433,8 @@ namespace stream
 		buf += 4; // reserve for lengthlength
 		msg->len += 4;
 
-		if (m_Gzip && m_Deflator)
-			size = m_Deflator->Deflate (payload, len, buf, msg->maxLen - msg->len);
+		if (m_Gzip || gzip)
+			size = m_Deflator.Deflate (payload, len, buf, msg->maxLen - msg->len);
 		else
 			size = i2p::data::GzipNoCompression (payload, len, buf, msg->maxLen - msg->len);
 
